from __future__ import annotations

import json
import os
from configparser import RawConfigParser
from pathlib import Path
from typing import Iterator, cast

import yaml

from dj_settings._seven import toml_parser
from dj_settings.constants import ETC, HOME_CONF, SUPPORTED_TYPES
from dj_settings.types import ConfDict, SupportedType


def get_override_paths(path: Path, *, same_suffix: bool) -> Iterator[Path]:
    if path.is_file() and os.access(path, os.R_OK):
        yield path

    suffix = path.suffix
    override_dir = path.with_suffix(f"{suffix}.d")
    if override_dir.is_dir():
        glob = f"*{suffix}" if same_suffix else f"{path.stem}*"
        for path in sorted(override_dir.glob(glob)):
            if path.is_file() and os.access(path, os.R_OK):
                yield path


def deep_merge(*dictionaries: ConfDict, merge_arrays: bool = False) -> ConfDict:
    output = dictionaries[0].copy()
    for dictionary in dictionaries[1:]:
        for key, value in dictionary.items():
            if isinstance(output.get(key), dict) and isinstance(value, dict):
                output[key] = deep_merge(output[key], value)
            elif (
                merge_arrays
                and isinstance(output.get(key), list)
                and isinstance(value, list)
            ):
                output[key] = output[key] + value
            else:
                output[key] = value
    return output


def _get_config_paths(filename: Path, project_dir: Path | None) -> Iterator[Path]:
    if project_dir is not None:
        yield project_dir.joinpath(filename)
    yield HOME_CONF.joinpath(filename)
    yield ETC.joinpath(filename)


def get_config_paths(
    filename: Path, *, project_dir: Path | None = None
) -> Iterator[Path]:
    for path in _get_config_paths(filename, project_dir=project_dir):
        if os.access(path, os.R_OK) and path.is_file():
            yield path


def get_type(path: Path, force_type: SupportedType | None = None) -> SupportedType:
    if force_type:
        if force_type not in SUPPORTED_TYPES:
            msg = f"{force_type} is not a supported extension (yet)"
            raise ValueError(msg)
        return force_type

    suffix = path.suffix
    if suffix in {".conf", ".cfg", ".ini"}:
        return "ini"
    if suffix in {".yml", ".yaml"}:
        return "yaml"
    if suffix == ".toml":
        return "toml"
    if suffix == ".json":
        return "json"
    if path.name.startswith(".env"):
        return "env"
    msg = f"Cannot infer type of {path}"
    raise ValueError(msg)


def extract_data(path: Path, settings_type: SupportedType) -> ConfDict:
    if settings_type == "env":
        data = {}
        with path.open() as file:
            for raw_line in file:
                line = raw_line.strip()
                if not line or line.startswith("#"):
                    continue
                key, value = line.split("=")
                data[key] = value
        return data

    if settings_type == "json":
        with path.open() as file:
            return cast(ConfDict, json.load(file))

    if settings_type == "ini":
        parser = RawConfigParser(default_section=None)  # type: ignore[call-overload]
        parser.optionxform = lambda option: option
        parser.read(path)
        return {key: dict(value) for key, value in parser.items() if key is not None}

    if settings_type == "toml":
        with path.open("rb") as binary_file:
            return toml_parser(binary_file)

    with path.open() as file:
        return cast(ConfDict, yaml.safe_load(file))



from __future__ import annotations

import os
from dataclasses import dataclass, field
from itertools import chain
from pathlib import Path
from typing import Any, Callable, Iterable

from dj_settings._seven import get_annotations
from dj_settings.exceptions import SectionError
from dj_settings.types import ConfDict, SupportedType
from dj_settings.utils import (
    deep_merge,
    extract_data,
    get_config_paths,
    get_override_paths,
    get_type,
)


class ConfigParser:
    __slots__ = ["_paths", "_data", "_merge_arrays"]

    def __init__(
        self,
        paths: Iterable[str | Path],
        force_type: SupportedType | None = None,
        *,
        merge_arrays: bool = False,
    ):
        self._paths = {Path(path): get_type(Path(path), force_type) for path in paths}
        self._data: ConfDict | None = None
        self._merge_arrays = merge_arrays

    @property
    def data(self) -> ConfDict:
        if self._data is None:
            self._data = {}
            for base_path, base_type in self._paths.items():
                same_suffix = base_type != "env"
                for path in get_override_paths(base_path, same_suffix=same_suffix):
                    self._data = deep_merge(
                        self._data,
                        extract_data(path, base_type),
                        merge_arrays=self._merge_arrays,
                    )
        return self._data

    def extract_value(self, name: str, sections: Iterable[Any]) -> Any:
        data = self.data
        for section in chain(sections, [name]):
            try:
                data = data[section]
            except (KeyError, AttributeError) as exc:  # noqa: PERF203
                msg = "Missing section"
                raise SectionError(msg) from exc

        return data


def get_setting(
    name: str,
    *,
    use_env: bool | str = True,
    project_dir: str | Path | None = None,
    filename: str | Path | None = None,
    sections: Iterable[Any] = (),
    merge_arrays: bool = False,
    rtype: type = str,
    default: Any = None,
) -> Any:
    if use_env:
        env_var = name if use_env is True else use_env
        if os.getenv(env_var) is not None:
            return rtype(os.environ[env_var])

    if filename is not None:
        if project_dir is not None:
            project_dir = Path(project_dir)
        parser = ConfigParser(
            get_config_paths(Path(filename), project_dir=project_dir),
            merge_arrays=merge_arrays,
        )
        try:
            value = parser.extract_value(name, sections)
        except SectionError:
            pass
        else:
            return rtype(value)

    return default


class _SettingsField:
    __slots__ = ["name", "use_env", "sections", "merge_arrays", "rtype", "default"]

    def __init__(
        self,
        name: str,
        *,
        use_env: bool | str,
        sections: Iterable[Any],
        merge_arrays: bool,
        rtype: type,
        default: Any,
    ):
        self.name = name
        self.use_env = use_env
        self.sections = sections
        self.merge_arrays = merge_arrays
        self.rtype = rtype
        self.default = default

    def __call__(
        self, project_dir: Path | str | None, filename: Path | str | None
    ) -> Any:
        return get_setting(
            self.name,
            use_env=self.use_env,
            project_dir=project_dir,
            filename=filename,
            sections=self.sections,
            merge_arrays=self.merge_arrays,
            rtype=self.rtype,
            default=self.default,
        )


def config_value(
    name: str,
    *,
    use_env: bool | str = True,
    sections: Iterable[Any] = (),
    merge_arrays: bool = False,
    rtype: type = str,
    default: Any = None,
) -> Any:

    return _SettingsField(
        name,
        use_env=use_env,
        sections=sections,
        merge_arrays=merge_arrays,
        rtype=rtype,
        default=default,
    )


def _preprocess_class(
    cls: type, project_dir: Path | str | None, filename: Path | str | None
) -> type:
    for attribute in get_annotations(cls):
        value = getattr(cls, attribute, None)
        if isinstance(value, _SettingsField):
            setattr(cls, attribute, field(default=value(project_dir, filename)))
    return cls


def settings_class(
    project_dir: Path | str | None = None, filename: Path | str | None = None
) -> Callable[[type], type]:
    def wrap(cls: type) -> type:
        cls = _preprocess_class(cls, project_dir, filename)
        return dataclass(frozen=True)(cls)

    return wrap



from __future__ import annotations

import os
from pathlib import Path
from typing import get_args

from dj_settings.types import SupportedType

ETC = Path("/etc/")
HOME_CONF = Path(os.environ.get("XDG_CONFIG_HOME", Path.home().joinpath(".config/")))
SUPPORTED_TYPES = frozenset(get_args(SupportedType))



from __future__ import annotations

import sys

MINOR = sys.version_info.minor

if MINOR >= 10:
    import inspect

    get_annotations = inspect.get_annotations

else:
    from typing import Any

    def get_annotations(cls: type) -> dict[str, Any]:
        return cls.__annotations__


if MINOR >= 11:
    import tomllib

    toml_parser = tomllib.load
else:
    import tomli

    toml_parser = tomli.load



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Literal, Union

ConfDict = Dict[str, Any]
PathConf = Union[str, Path, ConfDict]
SupportedType = Literal["json", "env", "yaml", "ini", "toml"]
